package com.nsftools.zip;

/*
Copyright 2005 Julian Robichaux

Licensed under the Academic Free License version 2.1 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://opensource.org/licenses/afl-2.1.php

Except as expressly stated in the License, software distributed 
under the License is distributed on an "AS IS" BASIS, and WITHOUT 
WARRANTY, either express or implied. See the License for the specific 
language governing permissions and limitations under the License.

I chose this license because it seemed like a good mix between
"You should say where this code came from if you use it" and
"You can use it in any way you want".
*/


import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.text.DateFormat;

/**
 * ZipFileManager is a Java class that is meant to simplify the
 * process of reading from, creating, and modifying zip files.
 * <p>
 * Here's some example code of the class in action:
 * <p><blockquote><code>
 * 	// if this file doesn't exist, it will be created; otherwise, the existing file can be<br>
 * 	// read and updated<br>
 * 	ZipFileManager zipFile = new ZipFileManager("c:\\test.zip");<br>
 * 	<br>
 * 	// optionally redirect debug messages to System.out, in case we want to<br> 
 * 	// read them<br>
 * 	zipFile.setDebugStream(System.out);<br>
 * 	<br>
 * 	// optionally set a relative path, so any files we add that are beneath this<br>
 * 	// subdirectory will have the relative directory information stripped off<br>
 * 	// (for example, if the relative path is "c:\\temp" and the file is at "c:\\temp\\foo.txt"<br>
 * 	// it will be added as "foo.txt" instead of "temp\\foo.txt"; if the file is at "c:\\a1\\foo.txt"<br>
 * 	// the relative path will be ignored and it will be added as "a1\\foo.txt")<br>
 * 	zipFile.setRelativePath("c:\\temp");<br>
 * 	<br>
 * 	// if an entry with a label of "foo.txt" already exists in the file, remove it (an array<br>
 * 	// of the existing entries, if any, can be retrieved with a call to getEntryNames)<br>
 * 	zipFile.removeEntry("foo.txt");<br>
 * 	<br>
 * 	// add the file at "c:\\test.txt" to the zip file<br>
 * 	zipFile.addEntry("c:\\test.txt");<br>
 * 	<br>
 * 	// add the file at "c:\\test2.txt" only if the file on disc is newer than the existing file<br>
 * 	// in the zip file (or if the file is not yet in the zip file)<br>
 * 	zipFile.updateEntry("c:\\test2.txt");<br>
 * 	<br>
 * 	// add all files in the directory "c:\\temp\\a1" to the zip file, including subdirectories<br>
 * 	zipFile.addDirectoryEntry("c:\\temp\\a1");<br>
 * 	<br>
 * 	// update the contents of (or create) the zip file with your changes and close it<br>
 * 	// (you could also have called the processChanges method to update the<br>
 * 	// file and continue modifying it)<br>
 * 	zipFile.close();<br>
 * 	<br>
 * </code></blockquote>
 * <p>
 * version 1.0<br>
 * December 26, 2005<br>
 * Copyright 2005 Julian Robichaux -- http://www.nsftools.com
 * <p>
 * Licensed under the Academic Free License version 2.1 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * <p>
 * http://opensource.org/licenses/afl-2.1.php
 * 
 * @version 1.0
 * @author Julian Robichaux -- http://www.nsftools.com
 */

public class ZipFileManager {
	
	private ZipFile oldZip;
	private Hashtable entryList;
	private Hashtable newEntryList;
	private Hashtable newFilesAdded;
	private String newZipName = "";
	private OutputStream debugOut = null;
	private boolean changesPending = false;
	private String relativePath = "";
	
	
	/**
	  * Create a new ZipFileManager object, using the given zipFile as
	  * the file we'll be reading from and potentially updating. If the given
	  * file does not yet exist (i.e. you want to add entries to a new file), it
	  * will be created when you call processChanges or close. If the file
	  * exists and it's not a valid java.util.zip.ZipFile, you will get a ZipException
	  * and your new ZipFileManager object will not be instantiated.
	  */
	public ZipFileManager (File zipFile) 
			throws IOException {
		if (zipFile.exists())
			oldZip = new ZipFile(zipFile);
		else
			oldZip = null;
		
		// by default, we'll be overwriting the existing file if any changes are made
		newZipName = zipFile.getCanonicalPath();
		newFilesAdded = new Hashtable();
		resetInternals();
	}
	
	
	/**
	  * Same as the other constructor, but allows a file name String instead of
	  * a File object.
	  */
	public ZipFileManager (String zipFileName) 
			throws IOException {
		this(new File(zipFileName));
	}
	
	
	/**
	  * Private method that resets the global variables we use to track changes
	  * to the zip file.
	  */
	private void resetInternals () {
		entryList = new Hashtable();
		newEntryList = new Hashtable();
		changesPending = false;
		
		if (oldZip != null) {
			Enumeration e = oldZip.entries();
			while (e.hasMoreElements()) {
				ZipEntry ze = (ZipEntry)e.nextElement();
				entryList.put(ze.getName(), ze);
			}
		}
		
	}
	
	
	/**
	  * If you want the debug messages to be piped to an OutputStream, set it here
	  */
	public void setDebugStream (OutputStream out) {
		debugOut = out;
	}
	
	
	/**
	  * convenience method for writing to the debug OutputStream, if there is one
	  */
	private void debugPrint (String msg) {
		if (debugOut != null) {
			try {
				String timeStamp = DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.LONG).format(new Date());
				debugOut.write((timeStamp + ": " + msg + "\n").getBytes());
			} catch (Exception ignored) {
			}
		}
	}
	
	
	/**
	  * Return the list of entries that are currently in the zip file (not including any
	  * files you've added since the last call to processChanges). This list is a
	  * Hashtable, with the entry name as the key (essentially a relative file name),
	  * and a ZipEntry object as the element.
	  */
	public Hashtable getEntries () {
		return entryList;
	}
	
	
	/**
	  * Return the ZipEntry associated with the given entryName, if any
	  */
	public ZipEntry getEntry (String entryName) {
		return (ZipEntry)entryList.get(entryName);
	}
	
	
	/**
	  * Return a String array of the entry names that are currently in the zip file
	  * (not including any files you've added since the last call to processChanges)
	  */
	public String[] getEntryNames () {
		return getKeyArray(entryList);
	}
	
	
	/**
	  * Return a String array of the files that will be added to the zip file
	  * (this list is reset every time you make a call to processChanges)
	  */
	public String[] getNewEntryNames () {
		return getKeyArray(newEntryList);
	}
	
	
	/**
	 * Return a String array of the new files that have already been added to
	 * the zip file as a result to ALL previous calls to processChanges
	 * (not including any files you've added since the last call to processChanges)
	 */
	public String[] getNewFilesAdded () {
		return getKeyArray(newFilesAdded);
	}
	
	
	/**
	  * Private convenience method to turn the keys in our hashlist to an array
	  * of Strings.
	  */
	private String[] getKeyArray (Hashtable ht) {
		if ((ht == null) || (ht.isEmpty())) {
			String[] keys = new String[1];
			keys[0] = "";
			return keys;
		}
		
		String[] keys = new String[ht.size()];
		int count = 0;
		
		for (Enumeration e = ht.keys() ; e.hasMoreElements() ;) {
			keys[count++] = (String)e.nextElement();
		}
		
		return keys;
	}
	
	
	
	/**
	  * Get the given entry as an InputStream, if it currently exists in the zip file. 
	  * To get the list of entries that are currently in the zip file, use getEntryNames.
	  */
	public InputStream getEntryStream (String entryName) 
			throws IOException {
		ZipEntry ze = (ZipEntry)entryList.get(entryName);
		if (ze != null)
			return oldZip.getInputStream(ze);
		else
			return null;
	}
	
	
	/**
	  * If the given entryName is found in the current zip file, decompress it and
	  * save it on disc to the given newFileName. The entry will NOT be removed
	  * from the zip file after it's saved to disc.
	  */
	public void saveEntryAsFile (String entryName, String newFileName)
			throws IOException {
		FileOutputStream out = new FileOutputStream(newFileName);
		InputStream in = getEntryStream(entryName);
		
		byte[] buf = new byte[10240];
		int buflen;
		while((buflen = in.read(buf)) >= 0)
			out.write(buf, 0, buflen);
		
		in.close();
		out.close();
	}
	
	
	/**
	  * Remove the specified entry from the current zip file, if it exists.
	  * The change will NOT take effect until you call processChanges
	  * or close.
	  */
	public void removeEntry (String entryName) {
		entryList.remove(entryName);
		changesPending = true;
		debugPrint("Removed " + entryName + " from old entry list");
	}
	
	
	/**
	  * Set the relative path for files that are added to this zip file. For example,
	  * if you set the relative path as "c:\temp" and you add a file at "c:\temp\foo.txt",
	  * it will be added as "foo.txt". If no relative path has been set (or the relative
	  * path is to a different directory), the file would be added as "temp\foo.txt".
	  * The relative path can be changed multiple times -- as new files are added,
	  * the most recent relative path assignment will be applied.
	  */
	public void setRelativePath (String pathName) {
		if (pathName != null)
			relativePath = pathName;
	}
	
	
	/**
	  * Get the current relative path assignment that is in effect.
	  */
	public String getRelativePath () {
		return relativePath;
	}
	
	
	/**
	  * Private method to convert a full file path to a relative one. Drive letter
	  * specifications (like "c:\") will always be removed, regardless of what
	  * the relative path is. The resulting String is used for the entry name in
	  * the new zip file.
	  */
	private String convertToRelativePath (String pathName) {
		return convertToRelativePath(pathName, relativePath);
	}
	
	
	/**
	  * Same as the other version of this method, only you can override the
	  * global relative path with a different pathname.
	  */
	private String convertToRelativePath (String pathName, String relPath) {
		try {
			if (pathName.toLowerCase().indexOf(relPath.toLowerCase()) == 0)
				pathName = pathName.substring(relPath.length());
			
			if (pathName.indexOf(":") > 0)
				pathName = pathName.substring(pathName.indexOf(":") + 1);
			while (pathName.charAt(0) == File.separatorChar)
				pathName = pathName.substring(1);
			
			if (pathName.toLowerCase().indexOf(relPath.toLowerCase()) == 0)
				pathName = pathName.substring(relPath.length());
		} catch (Exception ignored) {
		}
		
		return pathName;
	}
	
	
	/**
	  * Add the specified file to the new zip file, with the specified comment.
	  * The change will NOT take effect until you call processChanges
	  * or close.
	  * The maximum length of a comment is 0xFFFF (65,535) bytes, so any
	  * comment string longer than that will be truncated accordingly. This
	  * should not normally be an issue.
	  */
	public void addEntry (String fileName, String comment) 
			throws IllegalArgumentException {
		// exit early if we have bogus values
		if ((fileName == null) || fileName.equals(""))
			return;
		
		File checkFile = new File(fileName);
		if (!checkFile.exists()) {
			debugPrint("Cannot add " + fileName + " to new entry list: File does not exist.");
			return;
		} else if (checkFile.isDirectory()) {
			debugPrint("Cannot add " + fileName + " to new entry list: File is a directory.");
			return;
		}
		
		// relative path for the zip file entry
		String zname = convertToRelativePath(fileName);
		
		ZipEntry ze = new ZipEntry(zname);
		ze.setTime(checkFile.lastModified());
		
		if (comment != null) {
			if (comment.length() < 0xFFFF)
				ze.setComment(comment);
			else
				ze.setComment(comment.substring(0, 0xFFFF));
		}
		
		newEntryList.put(fileName, ze);
		
		changesPending = true;
		debugPrint("Added " + fileName + " to new entry list");
	}
	
	
	/**
	  * Add the specified file to the new zip file.
	  * The change will NOT take effect until you call processChanges
	  * or close.
	  */
	public void addEntry (String fileName) 
			throws IllegalArgumentException {
		addEntry(fileName, "");
	}
	
	
	/**
	  * Add the specified file to the new zip file.
	  * The change will NOT take effect until you call processChanges
	  * or close.
	  */
	public void addEntry (File file) 
			throws IllegalArgumentException, IOException {
		addEntry(file.getCanonicalPath(), "");
	}
	
	
	/**
	  * Add all files in the specified directory to the new zip file, including
	  * all subdirectories.
	  * The change will NOT take effect until you call processChanges or close.
	  */
	public void addDirectoryEntry (String dirName) 
			throws IllegalArgumentException, IOException {
		addDirectoryEntry(new File(dirName));
	}
	
	
	/**
	  * Add all files in the specified directory to the new zip file, including
	  * all subdirectories.
	  * The change will NOT take effect until you call processChanges or close.
	  */
	public void addDirectoryEntry (File dir) 
			throws IllegalArgumentException, IOException {
		addDirectoryEntry(dir, -1, null);
	}
	
	
	/**
	  * Add all files in the specified directory to the new zip file. The howDeep
	  * parameter determines how many levels of subdirectories we should
	  * traverse. Some examples:
	  *   If howDeep = 0, no subdirectories will be added (only the files at the
	  *        root of the base directory are added)
	  *   If howDeep = 1, the first level of subdirectories will be added
	  *   If howDeep = -1, all subdirectories will be added.
	  * The change will NOT take effect until you call processChanges or close.
	  */
	public void addDirectoryEntry (String dirName, int howDeep) 
			throws IllegalArgumentException, IOException {
		addDirectoryEntry(new File(dirName), howDeep);
	}
	
	
	/**
	  * Add all files in the specified directory to the new zip file. The howDeep
	  * parameter determines how many levels of subdirectories we should
	  * traverse. Some examples:
	  *   If howDeep = 0, no subdirectories will be added (only the files at the
	  *        root of the base directory are added)
	  *   If howDeep = 1, the first level of subdirectories will be added
	  *   If howDeep = -1, all subdirectories will be added.
	  * The change will NOT take effect until you call processChanges or close.
	  */
	public void addDirectoryEntry (File dir, int howDeep) 
			throws IllegalArgumentException, IOException {
		addDirectoryEntry(dir, howDeep, null);
	}
	
	
	/**
	  * Add all files in the specified directory to the new zip file. The howDeep
	  * parameter determines how many levels of subdirectories we should
	  * traverse. Some examples:
	  *   If howDeep = 0, no subdirectories will be added (only the files at the
	  *        root of the base directory are added)
	  *   If howDeep = 1, the first level of subdirectories will be added
	  *   If howDeep = -1, all subdirectories will be added.
	  * A FileFilter can also be specified, to limit the files that are added 
	  * to a certain subset of files (like only .log files, or everything but .sys files, 
	  * or something similar).
	  * The change will NOT take effect until you call processChanges or close.
	  */
	public void addDirectoryEntry (File dir, int howDeep, FileFilter filter) 
			throws IllegalArgumentException, IOException {
		if (dir.isDirectory()) {
			File[] children = dir.listFiles(filter);
			for (int i = 0; i < children.length; i++) {
				File child = children[i];
				if ((child.isDirectory()) && (howDeep != 0)) {
					addDirectoryEntry(child, (howDeep - 1), filter);
				} else {
					addEntry(child);
				}
			}
		}
		
	}
	
	
	/**
	  * Compare the given entry to the given file, and if the file on disc is
	  * newer than the entry in the zip file, replace the entry with the file.
	  * If the entry does not already exist in the zip file, the file will just be
	  * added. If it exists but the last modified time is the same as or after
	  * the modified time of the file on disc, nothing will be changed.
	  * The change will NOT take effect until you call processChanges
	  * or close.
	  */
	public void updateEntry (String entryName, File file) 
			throws IOException {
		ZipEntry ze = (ZipEntry)entryList.get(entryName);
		if (ze == null) {
			addEntry(file);
		} else { 
			if (file.lastModified() > ze.getTime()) {
				addEntry(file.getCanonicalPath(), ze.getComment());
				removeEntry(entryName);
			} else {
				debugPrint(entryName + " in zip file was not older than file on disc. No update.");
			}
		}
	}
	
	
	/**
	  * Same as the other updateEntry method, only you can use a fileName
	  * String instead of a File object.
	  */
	public void updateEntry (String entryName, String fileName) 
			throws IOException {
		updateEntry(entryName, new File(fileName));
	}
	
	
	/**
	  * If this file already exists in the zip file (based on the relative path name
	  * of the file) and the version on disc is newer than the version in the zip
	  * file, replace the version in the zip file with the one on disc. If the file does
	  * not already exist in the zip file (based on relative path) or the version in
	  * the zip file was modified at the same time or after the version on disc,
	  * nothing will happen.
	  * The change will NOT take effect until you call processChanges
	  * or close.
	  */
	public void updateEntry (File file) 
			throws IOException {
		updateEntry(convertToRelativePath(file.getCanonicalPath()), file);
	}
	
	
	/**
	  * Same as the other updateEntry method, only you can use a fileName
	  * String instead of a File object.
	  */
	public void updateEntry (String fileName) 
			throws IOException {
		updateEntry(convertToRelativePath(fileName), new File(fileName));
	}
	
	
	/**
	  * Same as addDirectoryEntry, only it uses the logic in the updateEntry
	  * method for each file instead of the logic in addEntry.
	  */
	public void updateDirectoryEntry (String dirName) 
			throws IllegalArgumentException, IOException {
		updateDirectoryEntry(new File(dirName));
	}
	
	
	/**
	  * Same as addDirectoryEntry, only it uses the logic in the updateEntry
	  * method for each file instead of the logic in addEntry.
	  */
	public void updateDirectoryEntry (File dir) 
			throws IllegalArgumentException, IOException {
		updateDirectoryEntry(dir, -1, null);
	}
	
	
	/**
	  * Same as addDirectoryEntry, only it uses the logic in the updateEntry
	  * method for each file instead of the logic in addEntry.
	  */
	public void updateDirectoryEntry (String dirName, int howDeep) 
			throws IllegalArgumentException, IOException {
		updateDirectoryEntry(new File(dirName), howDeep);
	}
	
	
	/**
	  * Same as addDirectoryEntry, only it uses the logic in the updateEntry
	  * method for each file instead of the logic in addEntry.
	  */
	public void updateDirectoryEntry (File dir, int howDeep) 
			throws IllegalArgumentException, IOException {
		updateDirectoryEntry(dir, howDeep, null);
	}
	
	
	/**
	  * Same as addDirectoryEntry, only it uses the logic in the updateEntry
	  * method for each file instead of the logic in addEntry.
	  */
	public void updateDirectoryEntry (File dir, int howDeep, FileFilter filter) 
			throws IllegalArgumentException, IOException {
		if (dir.isDirectory()) {
			File[] children = dir.listFiles(filter);
			for (int i = 0; i < children.length; i++) {
				File child = children[i];
				if (child.isDirectory() && (howDeep != 0)) {
					updateDirectoryEntry(child, (howDeep - 1), filter);
				} else {
					updateEntry(child);
				}
			}
		}
		
	}
	
	
	/**
	  * Return the name of the zip file we're currently working with, if any. If it's a
	  * new file that will be created when processChanges is called, this will return
	  * an empty String (""); however, you could get the name that the new file will
	  * be called with the getNewZipFileName method. If you've already made a
	  * call to processChanges and this value is NOT the same as getNewZipFileName,
	  * that likely means that there was an error renaming the temp file. In that case,
	  * this will be the name of the temp file that was created as the zip file was
	  * being created.
	  */
	public String getCurrentZipFileName () {
		if (oldZip == null)
			return "";
		else
			return oldZip.getName();
	}
	
	
	/**
	  * Return the name of the zip file that will be created when processChanges or
	  * close is called. By default, this is the same as the name of the file that was
	  * specified when this ZipFileManager object was created, which means any
	  * original file will be overwritten. 
	  */
	public String getNewZipFileName () {
		return newZipName;
	}
	
	
	/**
	  * Change the name of the zip file we will output our changes to. By default, we
	  * overwrite the file that was specified when this ZipFileManager object was
	  * created. However, you may want to leave the original file intact and make
	  * a brand new file based on the original file, in which case you would specify
	  * a new name here. If the file name you specify is invalid, no error will occur
	  * until you call processChanges or close, at which time the zip file will be
	  * created, but it will be in a temporary file instead of the file name you want.
	  */
	public void setNewZipFileName (String newFileName) {
		if ((newFileName != null) && (!newFileName.equals(newZipName))) {
			newZipName = newFileName;
			changesPending = true;
		}
	}
	
	
	/**
	  * Same as the other version of processChanges, with ignoreErrors
	  * set to false (which means the process will fail with an Exception 
	  * if there are I/O errors reading or writing files).
	  */
	public String processChanges () 
			throws IOException {
		return processChanges(false);
	}
	
	
	/**
	  * Process all outstanding changes to the zip file. This includes
	  * removing entries, adding or updating existing entries, or creating
	  * the new zip file (if the file does not yet exist, or if the newZipFileName
	  * is different from the current zip file name).
	  *
	  * With a zip file in Java, you can't make updates to an existing zip file
	  * in-place, so what you have to do is copy all the old entries you want to
	  * keep into a new file, add any new entries to that file, close the file stream
	  * and copy the new file over the old one. Normally this is pretty fast, but it
	  * obviously slows down as the files increase in size and/or number.
	  * 
	  * The ignoreErrors parameter indicates whether IO errors that occur as
	  * we're writing to the new zip file should be ignored or not. Errors could
	  * happen if a filename is incorrect, or the user doesn't have access to
	  * open the file, for example. If ignoreErrors is true, the errors will be ignored
	  * and the new zip file will be created without the files that caused errors
	  * (error messages will show up on the debug stream, if you're interested).
	  * If ignoreErrors is false, the entire process will throw an IOException at the
	  * first error and the new zip file will NOT be created.
	  * 
	  * If this method completes successfully, the return value will be the file name
	  * and path of the zip file that was created. This will either be the value of the
	  * newZipFileName variable (which is, by default, the name of the zip file that
	  * was specified when this ZipFileManager object was created), or if there was
	  * an error writing to the location specified by newZipFileName (often an error
	  * regarding write or delete permissions), the name of the temporary file that
	  * contains the new zip file contents.
	  */
	public String processChanges (boolean ignoreErrors) 
			throws IOException {
		if (!changesPending) {
			debugPrint("There are no pending changes. No processing will be done.");
			return newZipName;
		}
		
		debugPrint("Processing changes...");
		File zipFile = File.createTempFile("ZFM_", ".zip");
		debugPrint("Created temp file " + zipFile.getCanonicalPath());
		
		ZipOutputStream zipStream = new ZipOutputStream(new FileOutputStream(zipFile));
		Hashtable filesAdded = new Hashtable();
		
		// first add all the existing files that haven't been marked for removal
		for (Enumeration e = entryList.elements() ; e.hasMoreElements() ;) {
			ZipEntry ze = (ZipEntry)e.nextElement();
			debugPrint("Re-adding old entry " + ze.getName());
			AddEntry(zipStream, ze, oldZip.getInputStream(ze), ignoreErrors);
		}
		
		// then add any new files that were specified
		for (Enumeration e = newEntryList.keys() ; e.hasMoreElements() ;) {
			String fileName = (String)e.nextElement();
			ZipEntry ze = (ZipEntry)newEntryList.get(fileName);
			debugPrint("Adding new entry " + fileName + " as " + ze.getName());
			boolean success = AddEntry(zipStream, ze, new FileInputStream(fileName), ignoreErrors);
			if (success)
				filesAdded.put(fileName, ze);
		}
		
		// close the file stream to save the new zip file (still in the temp location)
		debugPrint("Finished adding entries. Closing file stream.");
		zipStream.close();
		
		// if this worked, it's okay to update our newFilesAdded list
		newFilesAdded.putAll(filesAdded);
		
		// now let's try to rename the temp file. If this fails, we won't explicitly
		// throw an error but we'll report the error to the debug stream and
		// this method will return the name of the temp file instead of the name
		// that the user wanted
		try {
			debugPrint("Attempting to rename temp file to " + newZipName);
			if (oldZip != null)
				oldZip.close();
			
			File newFile = new File(newZipName);
			if (newFile.exists())
				newFile.delete();
			
			if (zipFile.renameTo(newFile)) {
				oldZip = new ZipFile(newZipName);
				zipFile.deleteOnExit();
				debugPrint("Rename was successful.");
			} else {
				oldZip = new ZipFile(zipFile);
				debugPrint("Rename was NOT successful.");
				throw new IOException(zipFile.getCanonicalPath() + " could not be renamed to " +
						newZipName + ". Make sure path and permissions are correct");
			}
			
			resetInternals();
		} catch (IOException e) {
			debugPrint("ERROR renaming temp file: " + e.getMessage());
			throw e;
		}
		
		return zipFile.getName();
	}
	
	
	/**
	  * Private method to add a file to the new zip file we're creating
	  */
	private boolean AddEntry (ZipOutputStream zipStream, ZipEntry ze, InputStream in, boolean ignoreErrors) 
			throws IOException {
		boolean returnVal = false;
		
		try {
			zipStream.putNextEntry(ze);
			
			// read the file and send it to the ZipOutputStream
			byte[] buf = new byte[10240];
			int buflen;
			while((buflen = in.read(buf)) >= 0)
				zipStream.write(buf, 0, buflen);
			
			// close the entry and the stream
			zipStream.closeEntry();
			in.close();
			returnVal = true;
		} catch (IOException e) {
			debugPrint("ERROR adding entry " + ze.getName() + ": " + e.getMessage());
			if (!ignoreErrors)
				throw e;
		}
		
		return returnVal;
	}
	
	
	/**
	  * Returns true if there are changes pending (adds, removes, or updates)
	  * since the last time processChanges was called.
	  */
	public boolean areChangesPending () {
		return changesPending;
	}
	
	
	/**
	  * Write any outstanding changes (adds, removes, or updates) to the
	  * zip file (if any) and close the internal reference to the zip file.
	  */
	public void close () 
			throws IOException {
		close(true);
	}
	
	
	/**
	  * Close the internal reference to the zip file. If doProcessChanges is true and
	  * there are outstanding changes to be made (adds, removes, or updates),
	  * the processChanges method will be called first.
	  */
	public void close (boolean doProcessChanges) 
			throws IOException {
		try {
			if (changesPending && doProcessChanges)
				processChanges();
		} catch (IOException e) {
			throw e;
		} finally {
			if (oldZip != null)
				oldZip.close();
		}
	}
	
}


